/**
 * WildCAT: A Generic Framework for Context-Aware Applications.
 * Copyright (C) 2009 Bull S.A.S.
 * Copyright (C) 2009 EMN
 * Copyright (C) 2009 IRISA
 * Contact: wildcat@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 */
package org.ow2.wildcat.test;

import org.ow2.wildcat.WAction;
import org.ow2.wildcat.Context;
import org.ow2.wildcat.ContextException;
import org.ow2.wildcat.ContextFactory;
import org.ow2.wildcat.Query;

import org.testng.Assert;
import org.testng.annotations.AfterTest;
import org.testng.annotations.BeforeTest;
import org.testng.annotations.Test;

/**
 * Unit tests of {@link WAction} related operations
 * @author Mahmoud Ben Hassine
 *
 */
public class ActionTest {

	private Context ctx;
	private Query query;
	
	@BeforeTest
	protected void setUp() throws Exception {
		ctx = ContextFactory.getDefaultFactory().createContext();
		query = ctx.createQuery("select * from WEvent");
	}

    /**
     * Test method for
     * {@link org.ow2.wildcat.Context#registerActions(Query, WAction...)}.
     * @throws ContextException
     */
	@Test
	public void testRegisterOneAction() throws ContextException{
		IncrementCounterAction a = new IncrementCounterAction();		
        ctx.createAttribute("self://r#a", "hello");
		ctx.registerActions(query, a);
		ctx.setValue("self://r#a", "world");//trigger an event corresponding to the associated query
		Assert.assertEquals(1, a.getCounter());//test if the associated action was performed
	}
	
    /**
     * Test method for
     * {@link org.ow2.wildcat.Context#registerActions(Query, WAction...)}.
     * @throws ContextException
     */
	@Test
	public void testRegisterManyActions() throws ContextException{
		
		IncrementCounterAction a1 = new IncrementCounterAction();
		IncrementCounterAction a2 = new IncrementCounterAction();
		
        ctx.createAttribute("self://r#a", "hello");
		ctx.registerActions(query, a1,a2);
		ctx.setValue("self://r#a", "world");
		Assert.assertEquals(1, a1.getCounter());
		Assert.assertEquals(1, a2.getCounter());
	}
	
    /**
     * Test method for
     * {@link org.ow2.wildcat.Context#removeActions(Query, WAction...)}.
     * @throws ContextException
     */
	@Test
	public void testRemoveAction() throws ContextException{
		IncrementCounterAction a = new IncrementCounterAction();		
        ctx.createAttribute("self://r#a", "hello");
		ctx.registerActions(query, a);
		ctx.setValue("self://r#a", "world");
		Assert.assertEquals(1, a.getCounter());
		ctx.removeActions(query, a);
		ctx.setValue("self://r#a", "hello wildcat");//trigger an event after the action removal
		Assert.assertEquals(1, a.getCounter());//the action should not be performed at this point
	}
	
    /**
     * Test method for
     * {@link org.ow2.wildcat.Context#removeActions(Query, WAction...)}.
     * @throws ContextException
     */
	@Test
	public void testRemoveAllActions() throws ContextException{
		
		IncrementCounterAction a1 = new IncrementCounterAction();
		IncrementCounterAction a2 = new IncrementCounterAction();
		
        ctx.createAttribute("self://r#a", "hello");
		ctx.registerActions(query, a1,a2);
		ctx.setValue("self://r#a", "world");
		Assert.assertEquals(1, a1.getCounter());
		Assert.assertEquals(1, a2.getCounter());
		ctx.removeActions(query, a1,a2);
		ctx.setValue("self://r#a", "hello wildcat");
		Assert.assertEquals(1, a1.getCounter());
		Assert.assertEquals(1, a2.getCounter());
	}
	
    /**
     * Test method for
     * {@link org.ow2.wildcat.Context#removeActions(Query, WAction...)}.
     * @throws ContextException
     */
	@Test
	public void testRemoveOnlyOneAction() throws ContextException{
		
		IncrementCounterAction a1 = new IncrementCounterAction();
		IncrementCounterAction a2 = new IncrementCounterAction();
		
        ctx.createAttribute("self://r#a", "hello");
		ctx.registerActions(query, a1,a2);
		ctx.setValue("self://r#a", "world");
		Assert.assertEquals(1, a1.getCounter());
		Assert.assertEquals(1, a2.getCounter());
		ctx.removeActions(query, a1);//remove only a1
		ctx.setValue("self://r#a", "hello wildcat");
		Assert.assertEquals(1, a1.getCounter());//a1 should not be performed 
		Assert.assertEquals(2, a2.getCounter());//a2 should be performed
	}
	
	@AfterTest
	protected void tearDown() throws Exception {
		ctx = null;
		query = null;
		System.gc();
	}

	
	/**
	 * Test specific class used to test if an action was performed when an event is triggered
	 */
	class IncrementCounterAction extends WAction{

		private int counter;
		
		public IncrementCounterAction() {
			counter = 0;
		}
		@Override
		public void onEvent() {
			counter++;
		}
		
		public int getCounter(){
			return counter;
		}
	}
}
